# -*- python -*-
# $RCSfile: microstructureIO.py,v $
# $Revision: 1.13 $
# $Author: langer $
# $Date: 2004/08/08 02:27:38 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from SWIG.common import pixelattribute
from common import debug
from common import primitives
from common.IO import mainmenu
from common.IO import oofmenu
from common.IO import parameter
from common.IO import whoville
import common.microstructure

OOFMenuItem = oofmenu.OOFMenuItem

OOF = mainmenu.OOF
micromenu = OOF.LoadData.addItem(OOFMenuItem('Microstructure'))

#########

def _newMicrostructure(menuitem, name, isize, size):
    try:
        oldms = common.microstructure.microStructures[name]
    except KeyError:
        pass
    else:
        oldms.getObject().destroy()
    ms = common.microstructure.Microstructure(name, isize, size)

micromenu.addItem(OOFMenuItem(
    'New',
    callback=_newMicrostructure,
    params=[parameter.StringParameter('name'),
            primitives.iPointParameter('isize'),
            primitives.PointParameter('size')],
    help="Create a new Microstructure.  Used internally in data files."))

##########    

def _newpixelgroup(menuitem, microstructure, group, meshable):
    mscontext = common.microstructure.microStructures[microstructure]
    ms = mscontext.getObject()
    mscontext.begin_writing()
    try:
        grp = ms.getGroup(group)
        grp.set_meshable(meshable)
    finally:
        mscontext.end_writing()

micromenu.addItem(oofmenu.OOFMenuItem('PixelGroup',
                                      callback=_newpixelgroup,
                                      params=[
    whoville.WhoParameter('microstructure',
                          common.microstructure.microStructures),
    parameter.StringParameter('group'),
    parameter.BooleanParameter('meshable')
    ],
                                      help="Create a pixel group.  Used internally in data files."))

##########

categorymenu = micromenu.addItem(oofmenu.OOFMenuItem('DefineCategory'))

class CategoryMap:
    def __init__(self, array):
        # array is a list of lists of Ints, which are pixel categories
        self.pxls = {}                  # list of pixels, keyed by category
        for j in range(len(array)):
            row = array[j]
            for i in range(len(row)):
                where = primitives.iPoint(i, j)
                category = row[i]
                try:
                    self.pxls[category].append(where)
                except KeyError:
                    self.pxls[category] = [where]
    def getPixels(self, category):
        try:
            return self.pxls[category]
        except KeyError:
            return []
            
categoryMap = {} # CategoryMaps keyed by microstructure name

def getCategoryPixels(microstructure, category):
    map = categoryMap[microstructure]
    return map.getPixels(category)
    

def _readCategories(menuitem, microstructure, categories):
    categoryMap[microstructure] = CategoryMap(categories)

micromenu.addItem(oofmenu.OOFMenuItem('Categories',
                                      callback=_readCategories,
                                      params=[
    whoville.WhoParameter('microstructure',
                          common.microstructure.microStructures),
    parameter.ListOfListOfIntsParameter('categories')],
                                      help="Assign pixels to categories. Used internally in data files."))

#########

def _endCategories(menuitem, microstructure):
    del categoryMap[microstructure]

micromenu.addItem(oofmenu.OOFMenuItem('EndCategories',
                                      callback=_endCategories,
                                      params=[
    whoville.WhoParameter('microstructure',
                          common.microstructure.microStructures)],
                                      help="Clean up after pixel category definition.  Used internally in data files."))
##########

def writeMicrostructure(datafile, mscontext):
    ms = mscontext.getObject()
    
    mscontext.begin_writing()           # see comment below about categories
    unmeshablegroups = []
    try:
        datafile.startCmd(OOF.LoadData.Microstructure.New)
        datafile.argument('name', ms.name())
        datafile.argument('size', ms.size())
        datafile.argument('isize', ms.sizeInPixels())
        datafile.endCmd()

        
        for image in ms.images:
            image.writeImage(datafile)

        # Store pixel attributes by storing category definitions.  Pixel
        # groups that aren't marked as "meshable" don't influence the
        # category computation, so we have to temporarily make all groups
        # meshable.  (This is why we need to acquire the Microstructure's
        # *write* lock.)

        for grpname in ms.groupNames():
            grp = ms.findGroup(grpname)
            meshable = grp.is_meshable()
            datafile.startCmd(OOF.LoadData.Microstructure.PixelGroup)
            datafile.argument('microstructure', ms.name())
            datafile.argument('group', grpname)
            datafile.argument('meshable', meshable)
            datafile.endCmd()
            # Temporarily make all groups meshable
            if not meshable:
                unmeshablegroups.append(grp)
                grp.suppress_change_announcements()
                grp.set_meshable(1)

        # Save categories
        datafile.startCmd(OOF.LoadData.Microstructure.Categories)
        datafile.argument('microstructure', ms.name())
        datafile.argument('categories', ms.getCategoryMap())
        datafile.endCmd()
        
        # Save definitions of pixel categories
        for i in range(pixelattribute.nAttributes()):
            reg = pixelattribute.getRegistration(i)
            reg.writeGlobalData(datafile, ms)
            for category in range(ms.nCategories()):
                reppxl = ms.getRepresentativePixel(category)
                datafile.startCmd(getattr(
                    OOF.LoadData.Microstructure.DefineCategory, reg.name()))
                datafile.argument('microstructure', ms.name())
                datafile.argument('category', category)
                if reg.writeData(datafile, ms, reppxl):
                    datafile.endCmd()
                else:
                    datafile.discardCmd()

        datafile.startCmd(OOF.LoadData.Microstructure.EndCategories)
        datafile.argument('microstructure', ms.name())
        datafile.endCmd()
                
    finally:
        # Restore meshability of groups
        for grp in unmeshablegroups:
            grp.set_meshable(0)
            grp.allow_change_announcements()

        mscontext.end_writing()
        

    
