// -*- C++ -*-
// $RCSfile: cpixelselection.C,v $
// $Revision: 1.13 $
// $Author: langer $
// $Date: 2004/05/07 19:48:05 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "common/activearea.h"
#include "common/cmicrostructure.h"
#include "common/cpixelselection.h"
#include "printvec.h"

CPixelSelection::CPixelSelection(const ICoord *pxlsize, const Coord *size,
				 CMicrostructure *ms)
  : pixgrp("", pxlsize, (CMicrostructure*)0),
    bitmap(size, pxlsize),
    size_(*size),
    isize_(*pxlsize),
    microstructure(ms)
{}

CPixelSelection::CPixelSelection(const CPixelSelection &other)
  : pixgrp("", &other.sizeInPixels(), (CMicrostructure*)0),
    bitmap(&other.size(), &other.sizeInPixels()),
    size_(other.size()),
    isize_(other.sizeInPixels()),
    microstructure(other.microstructure)
{
  pixgrp.copy(&other.pixgrp);
  bitmap.copy(&other.bitmap);
}

bool CPixelSelection::checkpixel(const ICoord *pixel) const {
  return bitmap.contains(pixel);
}

const ActiveArea *CPixelSelection::getActiveArea() const {
  return microstructure->getActiveArea();
}

const std::vector<ICoord> *CPixelSelection::getActivePixels() const {
  // This returns a new'd vector which must be deleted.  The active
  // pixels are the ones *not* set in the bitmap, which is why the
  // argument to getPixels is 0.
  return getActiveArea()->getBitmap()->getPixels(0);
}

void CPixelSelection::clear() {
  if(getActiveArea()->getOverride())
    clearWithoutCheck();
  else {
    const std::vector<ICoord> *activepxls = getActivePixels();
    pixgrp.remove(activepxls);
    bitmap.reset(activepxls);
    delete activepxls;
  }
  ++timestamp;
}

void CPixelSelection::clearWithoutCheck() {
  pixgrp.clear();
  bitmap.clear();
  ++timestamp;
}

void CPixelSelection::invert() {
  if(getActiveArea()->getOverride())
    invertWithoutCheck();
  else {
    const std::vector<ICoord> *activepxls = getActivePixels();
    bitmap.toggle(activepxls);
    pixgrp.clear();
    pixgrp.setFromBitmap(bitmap);
  }
  ++timestamp;
};

void CPixelSelection::invertWithoutCheck() {
  bitmap.invert();
  pixgrp.clear();
  pixgrp.setFromBitmap(bitmap);
  ++timestamp;
};

void CPixelSelection::select(PixelSelectionCourier *selection) {
  const ActiveArea *activearea = getActiveArea();
  if(activearea->getOverride())
    selectWithoutCheck(selection);
  else {
    std::vector<ICoord> okpix;
    selection->start();
    while(!selection->done()) {
      ICoord pixel = selection->currentPoint();
      if(checkpixel(&pixel) && activearea->isActive(&pixel)) {
	bitmap.set(&pixel);
	okpix.push_back(pixel);
      }
      selection->next();
    }
    pixgrp.add(&okpix);
  }
  ++timestamp;
}

void CPixelSelection::unselect(PixelSelectionCourier *selection) {
  const ActiveArea *activearea = getActiveArea();
  if(activearea->getOverride())
    unselectWithoutCheck(selection);
  else {
    std::vector<ICoord> *pixels = new std::vector<ICoord>;
    selection->start();
    while(!selection->done()) {
      ICoord pixel = selection->currentPoint();
      if(checkpixel(&pixel) && activearea->isActive(&pixel)) {
	bitmap.reset(&pixel);
	pixels->push_back(pixel);
      }
      selection->next();
    }
    pixgrp.remove(pixels);
    delete pixels;
  }
  ++timestamp;
}

void CPixelSelection::toggle(PixelSelectionCourier *selection) {
  const ActiveArea *activearea = getActiveArea();
  if(activearea->getOverride())
    toggleWithoutCheck(selection);
  else {
    selection->start();
    while(!selection->done()) {
      ICoord pixel = selection->currentPoint();
      if(checkpixel(&pixel) && activearea->isActive(&pixel)) {
	bitmap.toggle(&pixel);
      }
      selection->next();
    }
    pixgrp.clear();
    pixgrp.setFromBitmap(bitmap);
  }
  ++timestamp;
}

void CPixelSelection::selectSelected(PixelSelectionCourier *selection) {
  const ActiveArea *activearea = getActiveArea();
  pixgrp.clear();
  selection->start();
  while(!selection->done()) {
    ICoord pixel = selection->currentPoint();
    if(bitmap.get(&pixel))
      pixgrp.add(pixel);
    selection->next();
  }
  bitmap.clear();
  bitmap.set(pixgrp.members());
  ++timestamp;
}

void CPixelSelection::selectWithoutCheck(PixelSelectionCourier *selection) {
  std::vector<ICoord> okpix;
  selection->start();
  while(!selection->done()) {
    ICoord pixel = selection->currentPoint();
    if(checkpixel(&pixel)) {
      bitmap.set(&pixel);
      okpix.push_back(pixel);
    }
    selection->next();
  }
  pixgrp.add(&okpix);
  ++timestamp;
}

void CPixelSelection::unselectWithoutCheck(PixelSelectionCourier *selection) {
  std::vector<ICoord> *pixels = new std::vector<ICoord>;
  selection->start();
  while(!selection->done()) {
    ICoord pixel = selection->currentPoint();
    if(checkpixel(&pixel)) {
      bitmap.reset(&pixel);
      pixels->push_back(pixel);
    }
    selection->next();
  }
  pixgrp.remove(pixels);
  delete pixels;
  ++timestamp;
}

void CPixelSelection::toggleWithoutCheck(PixelSelectionCourier *selection) {
  selection->start();
  while(!selection->done()) {
    ICoord pixel = selection->currentPoint();
    if(checkpixel(&pixel)) {
      bitmap.toggle(&pixel);
    }
    selection->next();
  }
  pixgrp.clear();
  pixgrp.setFromBitmap(bitmap);
  ++timestamp;
}

bool CPixelSelection::isSelected(const ICoord *pixel) const {
  return bitmap.get(pixel);
}

const std::vector<ICoord> *CPixelSelection::members() const {
  return pixgrp.members();
}

int CPixelSelection::len() const {
  return pixgrp.len();
}

void CPixelSelection::setFromGroup(const PixelGroup *grp) {
  bitmap.set(grp->members());
  pixgrp.setFromBitmap(bitmap);
}
