# -*- python -*-
# $RCSfile: analyze.py,v $
# $Revision: 1.14 $
# $Author: haan $
# $Date: 2004/09/02 16:00:08 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Objects for performing operations on data.  These include writing
# data to a file, taking averages and standard deviations, and so
# forth.  The data to which this should be done is specified by means
# of an output.

## TODO LATER: Completely restructure the Analyze machinery.  The basic
## Analyze command will take four arguments, in addition to the Mesh:

##   Data: A list of Outputs. It should be possible to output Material
##   parameters, too.

##   Domain: Where the data will be evaluated. There are four
##   subclasses of Domain:
##      Point: sets of points.  The interface will list the named
##      point boundaries, named sets of points, and have a "New..."
##      item that will bring up a dialog box for entering a new named
##      set of points.
##      Line: cross sections.  There are named cross sections (from
##      the Mesh CS toolbox) and explicit cross sections (with
##      specified end points).  Data can also be plotted along edge
##      boundaries.
##      Area:  The whole mesh, or sets of elements.
##      Conditional:  All elements for which some criterion is active.

## Operation: How the data will be treated.  There are (at least)
## three subclasses:
##      RawOutput:  just print the values.
##      Statistics:  print mean and deviation.
##      Distribution: print a histogram.

## Destination: Where the data is going.  Message Window, File, and
## Plotter are the obvious choices.

## The exact behavior of the Analyze function will probably have to
## depend on a double dispatch mechanism between the Domain and
## Operation classes.  Averaging over points is different than
## averaging over Elements.

## The GUI will have panels for each of the above four arguments and a
## big 'OK' button that will call OOF.Analyze().



from common import primitives
from common import registeredclass
from common.IO import automatic
from common.IO import parameter
from common.IO import destinationparam
from engine.IO import meshcsparams
import math
import string


# Parent class for all "Data Operation" objects, i.e. things which do
# things to ScalarOutputs, like writing them gridwise to files,
# evaluating them along cross-sections, and so forth.

# In all the subclasses, the __call__ method provides the mesh
# object, output method, and a file handle to write to.

class DataOperation(registeredclass.RegisteredClass):
    registry = []

class GridOutput(DataOperation):
    def __init__(self, x_points, y_points):
        self.x_points = x_points
        self.y_points = y_points

    def __call__(self, meshobj, o_put, outfile):
        fe_mesh = meshobj.getObject()
        size = meshobj.size()

        for i in range(self.x_points):
            x = size[0]*float(i)/float(self.x_points-1)
            for j in range(self.y_points):
                y = size[1]*float(j)/float(self.y_points-1)
                p = primitives.Point(x,y)

                skel = meshobj.skeleton.enclosingElement(p)
                el = fe_mesh.getElement(skel.index)
                res = o_put.evaluate(fe_mesh, [el],
                                     [ [el.to_master(p[0],p[1])] ])
                print >> outfile, x, y, res[0]
        

registeredclass.Registration(
    "Grid", DataOperation, GridOutput, 1,
    params=[ parameter.IntParameter("x_points", 50),
             parameter.IntParameter("y_points", 50) ],
    tip="Output data at a regular grid of points covering the whole mesh.")



# Utility list-manipulating function -- basically does a dot product.
# Lists must be numeric and have the same length.
def _contract(lst1, lst2):
    return reduce(lambda x,y:x+y[0]*y[1],zip(lst1,lst2),0.0)


# TODO LATER: Area-fraction-based stats can maybe use the contouring
# code to make histograms and such-like.

class AverageOutput(DataOperation):
    def __init__(self, order):
        self.order = order
        
    # This one integrates the output over each element and normalizes
    # by the area.
    def __call__(self, meshobj, o_put, outfile):
        integralsum = 0.0
        areasum = 0.0
        fe_mesh = meshobj.getObject()
        for el in fe_mesh.element_iterator():
            gps = el.integration_points(self.order)
            pts = [x.mastercoord() for x in gps]
            wgts = [x.weight() for x in gps]
            
            integral = _contract(wgts, o_put.evaluate(fe_mesh, [el], [pts]))
            area = reduce(lambda x,y: x+y, wgts)

            integralsum += integral
            areasum += area
            
        result = integralsum/areasum

        print >> outfile, `result`

        
registeredclass.Registration(
    "Average", DataOperation, AverageOutput, 2,
    params=[parameter.IntParameter("order", 2,
                                   tip="Order of integration to use.") ],
    tip="Average the data over all elements in the mesh."
    )



# Standard deviation, square root of the integral of ( (x-<x>)^2 ),
# which is of course (<x^2>-<x>^2)^(1/2)
class StdDevOutput(DataOperation):
    def __init__(self, order):
        self.order = order

    # Standard deviation.
    def __call__(self, meshobj, o_put, outfile):
        integralsum = 0.0
        areasum = 0.0
        squaresum = 0.0

        fe_mesh = meshobj.getObject()

        # Make many lists...
        for el in fe_mesh.element_iterator():
            gps = el.integration_points(self.order)
            pts = map(lambda x: x.mastercoord(), gps)
            wgt = map(lambda x: x.weight(), gps)
            vls = o_put.evaluate(fe_mesh, [el], [pts])

            vls2 = map(lambda x: x*x, vls)

            integralsum += _contract(wgt,vls)
            squaresum += _contract(wgt,vls2)
            areasum += reduce(lambda x,y: x+y, wgt)

        variance = (squaresum/areasum) - math.pow(integralsum/areasum, 2)
        print >> outfile, `math.sqrt(abs(variance))`
        


registeredclass.Registration(
    "Standard Deviation", DataOperation, StdDevOutput, 3,
    params=[
    parameter.IntParameter(
    "order", 2, tip="Order of integration to use.") ],
    tip="Compute the standard deviation of the data over all elements in the mesh.")


class CrossSectionOutput(DataOperation):
    def __init__(self, cross_section, points, plot_x, plot_y, plot_s):
        self.cross_section = cross_section
        self.points = points
        self.plot_x = plot_x
        self.plot_y = plot_y
        self.plot_s = plot_s

    def __call__(self, meshobj, o_put, outfile):
        fe_mesh = meshobj.getObject()

        cs_obj = meshobj.getCrossSection(self.cross_section)
        s_pt = cs_obj.start
        e_pt = cs_obj.end

        delta = e_pt - s_pt
        length = math.sqrt(delta**2)

        headers = ["#"]
        if self.plot_x: headers.append("x")
        if self.plot_y: headers.append("y")
        if self.plot_s: headers.append("arclength")
        headers.append(o_put.short_repr())
        print >> outfile, string.join(headers, " ")
        
        for p in [
            s_pt + (float(i)/float(self.points-1.0))*delta
            for i in range(self.points) ]:

            skel = meshobj.skeleton.enclosingElement(p)
            el = fe_mesh.getElement(skel.index)

            res = o_put.evaluate(fe_mesh, [el],
                                 [ [ el.to_master(p[0],p[1]) ] ])

            vals = []
            if self.plot_x: vals.append(`p[0]`)
            if self.plot_y: vals.append(`p[1]`)
            if self.plot_s: vals.append(`math.sqrt((p-s_pt)**2)`)
            vals.append(`res[0]`)

            print >> outfile, string.join(vals, ' ')


registeredclass.Registration(
    "Cross Section", DataOperation, CrossSectionOutput, 4,
    params=[
    meshcsparams.MeshCrossSectionParameter(
    "cross_section", tip="The cross section to use"),
    parameter.IntParameter("points", value=50,
                           tip="The number of evaluation points."),
    parameter.BooleanParameter('plot_x', value=1,
                               tip="Include x coordinates in the output?"),
    parameter.BooleanParameter('plot_y', value=1,
                               tip="Include y coordinates in the output?"),
    parameter.BooleanParameter('plot_s', value=0,
                               tip="Include arc length coordinates in the output?")
    ],
    tip="Plot data along a cross section through the mesh.  Cross sections can be defined in the Mesh Cross Section toolbox in a graphics window.")
