# -*- python -*-
# $RCSfile: pinnodes.py,v $
# $Revision: 1.26 $
# $Author: haan $
# $Date: 2004/03/25 22:34:04 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from engine.IO import pinnodemodifier
from SWIG.common import switchboard
from common import debug
from common import toolbox
from common import utils
from common.IO import oofmenu
from common.IO import parameter

# Toolbox for pinning nodes.  Pinned nodes don't move during mesh
# modification operations.  They *can* move during equilibration --
# they're not boundary conditions.  Pinned nodes are treated much like
# Skeleton Selectables, and share a lot of code with them.

class PinnedNodesToolbox(toolbox.Toolbox):
    def __init__(self, gfxwindow):
        toolbox.Toolbox.__init__(self, 'Pin_Nodes', gfxwindow)
        self.pinnednodes = None         # PinnedNodeSelection object
        self.sb_callbacks = [
            switchboard.requestCallback((self.gfxwindow(), 'layers changed'),
                                        self.refresh),
            switchboard.requestCallback(('who changed', 'Skeleton'),
                                        self.skelChanged),
            switchboard.requestCallback("new pinned nodes",
                                        self.newPinnedNodes)
            ]

    def refresh(self):
        skeletoncontext = self.gfxwindow().topwho('Skeleton')
        if skeletoncontext is not None:
            self.pinnednodes = skeletoncontext.pinnednodes # PinnedNodeSelection
        else:
            self.pinnednodes = None

    def skelChanged(self, context):
        if context is self.getSkeletonContext():
            self.refresh()

    def activate(self):
        self.refresh()

    def makeMenu(self, menu):
        self.menu = menu
        menu.addItem(oofmenu.OOFMenuItem(
            'Pin',
            callback = self.pin,
            params=[parameter.FloatParameter('x', 0.0),
                    parameter.FloatParameter('y', 0.0)],
            help="Pin the node closest to the point x,y."
            ))
        menu.addItem(oofmenu.OOFMenuItem(
            'UnPin',
            callback = self.unpin,
            params=[parameter.FloatParameter('x'),
                    parameter.FloatParameter('y')],
            help="Unpin the node closest to the point x,y."
            ))
        menu.addItem(oofmenu.OOFMenuItem(
            'TogglePin',
            callback=self.togglepin,
            params=[parameter.FloatParameter('x'),
                    parameter.FloatParameter('y')],
            help="Toggle the pinnedness of the node closest to the point x,y."
            ))
        menu.addItem(oofmenu.OOFMenuItem(
            'UnPinAll',
            callback = self.unpinall,
            help="Unpin all nodes."
            ))
        menu.addItem(oofmenu.OOFMenuItem(
            'Invert',
            callback = self.invert,
            help="Invert pinned nodes."
            ))        
        for registration in pinnodemodifier.PinNodeModifier.registry:
            menuitem = menu.addItem(
                oofmenu.OOFMenuItem(
                utils.space2underscore(registration.name()),
                callback=pinnodemodifier.pinnodemod,
                params=[parameter.StringParameter('context')]+
                registration.params))
            menuitem.data = registration

        menu.addItem(oofmenu.OOFMenuItem('Undo', callback=self.undoPin))
        menu.addItem(oofmenu.OOFMenuItem('Redo', callback=self.redoPin))

    def getSkeletonContext(self):
        if self.pinnednodes is not None:
            return self.pinnednodes.skeletoncontext
        else:
            return None
        
    def getSkeleton(self):
        if self.pinnednodes is not None:
            return self.pinnednodes.skeletoncontext.getObject()

    def pin(self, menuitem, x, y):
        if self.pinnednodes is not None:
            self.pinnednodes.start()
            self.pinnednodes.pinPoint(x, y)
            self.pinnednodes.signal()

    def unpin(self, menuitem, x, y):
        if self.pinnednodes is not None:
            self.pinnednodes.start()
            self.pinnednodes.unpinPoint(x, y)
            self.pinnednodes.signal()

    def togglepin(self, menuitem, x, y):
        if self.pinnednodes is not None:
            self.pinnednodes.start()
            self.pinnednodes.togglepinPoint(x, y)
            self.pinnednodes.signal()

    def notPinned(self, node):
        if node.pinned():
            return 0
        return 1
    
    def invert(self, menuitem):
        newpinned = filter(self.notPinned,
                           self.getSkeletonContext().getObject().node_iterator())
        self.pinnednodes.start()
        self.pinnednodes.clear()
        self.pinnednodes.pin(newpinned)
        self.pinnednodes.signal()

    def unpinall(self, menuitem):
        if self.pinnednodes is not None:
            self.pinnednodes.start()
            self.pinnednodes.clear()
            self.pinnednodes.signal()

##     def pinSelection(self, menuitem):
##         if self.pinnednodes is not None:
##             self.pinnednodes.start()
##             self.pinnednodes.pin(
##                 self.getSkeletonContext().nodeselection.retrieve())
##             self.pinnednodes.signal()

##     def unpinSelection(self, menuitem):
##         if self.pinnednodes is not None:
##             self.pinnednodes.start()
##             self.pinnednodes.unpin(
##                 self.getSkeletonContext().nodeselection.retrieve())
##             self.pinnednodes.signal()

    def undoPin(self, menuitem):
        if self.pinnednodes is not None:
            self.pinnednodes.undo()
            self.pinnednodes.signal()

    def redoPin(self, menuitem):
        if self.pinnednodes is not None:
            self.pinnednodes.redo()
            self.pinnednodes.signal()
            
    # switchboard "new pinned nodes"
    def newPinnedNodes(self, pinnednodesSelection):
        switchboard.notify((self.gfxwindow(), "new nodes pinned"),
                           pinnednodesSelection)
        switchboard.notify("redraw")

toolbox.registerToolboxClass(PinnedNodesToolbox, ordering=2.7)
