# -*- python -*-
# $RCSfile: rationalizefunctions.py,v $
# $Revision: 1.2 $
# $Author: haan $
# $Date: 2004/03/04 21:24:05 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from common import debug
from engine import skeleton
from engine import skeletonelement

ProvisionalChanges = skeleton.ProvisionalChanges
ProvisionalTriangle = skeletonelement.ProvisionalTriangle
ProvisionalQuad = skeletonelement.ProvisionalQuad


def removeShortSide(skel, element, which):
    node0 = element.nodes[which]
    node1 = element.nodes[(which+1)%4]
    return [skel.mergeNode(node0, node1),
            skel.mergeNode(node1, node0)]

#############################################################

def quadSplit(skel, element, which):
    nodes = element.nodes
    parents = element.getParents()
    if which%2 == 0:
        triangles = (ProvisionalTriangle([nodes[0], nodes[1], nodes[2]],
                                         parents=parents),
                     ProvisionalTriangle([nodes[0], nodes[2], nodes[3]],
                                         parents=parents))
    else:
        triangles = (ProvisionalTriangle([nodes[1], nodes[2], nodes[3]],
                                         parents=parents),
                     ProvisionalTriangle([nodes[0], nodes[1], nodes[3]],
                                         parents=parents))
    change = ProvisionalChanges()
    change.removeElements(element)
    change.insertElements(*triangles)
    return [change]

#############################################################

def removeBadTriangle(skel, element, obtuse_angle, acute_angles):        
    return obtuseHandler(skel, element, obtuse_angle) + \
           acuteHandler(skel, element, acute_angles[0]) + \
           acuteHandler(skel, element, acute_angles[1])
        
def acuteHandler(skel, element, index):
    # Merge the two nodes opposite the sharp angle.
    node0 = element.nodes[(index+1)%3]
    node1 = element.nodes[(index+2)%3]
    return [skel.mergeNode(node0, node1),
            skel.mergeNode(node1, node0)]

def obtuseHandler(skel, element, index):
    # A obute triangle. What we do depends on
    # what kind of element borders the side joining the
    # two sharp angles.
    anchor = element.nodes[index] # wide corner
    itchy = element.nodes[(index+1)%3] # sharp corner
    scratchy = element.nodes[(index+2)%3] # sharp corner
    borderelement = element.getSister(skel, itchy, scratchy)
    if borderelement is None:
        return triNoneSplit(skel, anchor, itchy, scratchy,
                            element)
    if borderelement.nnodes() == 3:
        return triTriSplit(skel, anchor, itchy, scratchy,
                           element, borderelement)
    if borderelement.nnodes() == 4:
        return triQuadSplit(skel, anchor, itchy, scratchy,
                            element, borderelement)
    return []        

def triQuadSplit(skel, anchor, itchy, scratchy, tri, quad):

    # If tri.dominantPixel is different from quad.dominantPixel ...
    # Pick the best of three configurations.
    #            ___________        ______      ______      ______
    # scratchy /|          Q1       |   /|      |   /|      |    |
    #         / |          |        |C / |      |B / |      |    |
    #        /  |          |        | /  |      | /  |      |  B |
    #       /tri|  quad    |  ===>  |/   |      |/   |      |    |
    # anchor\   |          |        |\ B |      |    |      |\   |
    #        \  |          |        | \  |      | A  |      | \  |
    #         \ |          |        |A \ |      |    |      |A \ |
    #   itchy  \|__________Q0       |___\|      |____|      |___\|
    #

    # If tri & quad have same dominantPixel ... pick the best one as the case above.
    #            ___________            _________
    #          /|          |           /      . |
    #         / |          |          / C  .    |
    #        /  |          |         /  .       | 
    #       /tri|  quad    |  ===>  /.     B    | Three differnt cases
    #       \   |          |        \   .       | as in the above
    #        \  |          |         \     .    |
    #         \ |          |          \  A    . |
    #          \|__________|           \________|
    #

    # Does ProvisionalChanges need to handle node motion too, in order
    # to compute the correct deltaE in case A?

    # If itchy and scratchy are pinned, the process should be aborted
    if itchy.pinned() and scratchy.pinned():
        return []
    
    # Find nodes Q0 and Q1
    quadnodes = quad.nodes
    itchyindex = quadnodes.index(itchy)
    q0 = quadnodes[(itchyindex+1)%4]
    q1 = quadnodes[(itchyindex+2)%4]

    # Cases for moving anchor
    midpoint = 0.5*(itchy.position() + scratchy.position())
    
    change0 = skeleton.ProvisionalChanges()
    change0.moveNode(anchor, midpoint)
    change0.removeElements(tri, quad)
    parents = quad.getParents()
    change0.insertElements(
        ProvisionalTriangle([anchor, itchy, q0], parents=parents),
        ProvisionalTriangle([anchor, q0, q1], parents=parents),
        ProvisionalTriangle([anchor, q1, scratchy], parents=parents))

    change1 = skeleton.ProvisionalChanges()
    change1.moveNode(anchor, midpoint)
    change1.removeElements(tri, quad)
    parents = quad.getParents()
    change1.insertElements(
        ProvisionalQuad([anchor, itchy, q0, q1], parents=parents),
        ProvisionalTriangle([anchor, q1, scratchy], parents=parents))

    change2 = skeleton.ProvisionalChanges()
    change2.moveNode(anchor, midpoint)
    change2.removeElements(tri, quad)
    parents = quad.getParents()
    change2.insertElements(
        ProvisionalTriangle([anchor, itchy, q0], parents=parents),
        ProvisionalQuad([anchor, q0, q1, scratchy], parents=parents))

    # Cases for lazy anchor
    change3 = skeleton.ProvisionalChanges()
    change3.removeElements(tri, quad)
    parents = quad.getParents() + tri.getParents()
    change3.insertElements(
        ProvisionalTriangle([anchor, itchy, q0], parents=parents),
        ProvisionalTriangle([anchor, q0, q1], parents=parents),
        ProvisionalTriangle([anchor, q1, scratchy], parents=parents))

    change4 = skeleton.ProvisionalChanges()
    change4.removeElements(tri, quad)
    parents = quad.getParents()
    change4.insertElements(
        ProvisionalQuad([anchor, itchy, q0, q1], parents=parents),
        ProvisionalTriangle([anchor, q1, scratchy], parents=parents))

    change5 = skeleton.ProvisionalChanges()
    change5.removeElements(tri, quad)
    parents = quad.getParents()
    change5.insertElements(
        ProvisionalTriangle([anchor, itchy, q0], parents=parents),
        ProvisionalQuad([anchor, q0, q1, scratchy], parents=parents))
    
    # If anchor is limited in movility, change0-case2 is not possible.
    if not anchor.movable_x() or not anchor.movable_y():
        return [change3, change4, change5]
    else:
        return [change0, change1, change2, change3, change4, change5]

###########################################################################

def triTriSplit(skel, anchor, itchy, scratchy, tri1, tri2):
    #
    #                                  
    # scratchy /|\               |\    
    #         / | \              | \   
    #        /  |  \           C | A\   If tri1.dominantPixel is different
    #       /tri|tri\T0    ===>  |___\  from tri2.dominantPixel ...
    # anchor\ 1 | 2 /            |   / 
    #        \  |  /           D | B/  
    #         \ | /              | /   
    #   itchy  \|/               |/    
    #
    #
    #          /|\              / \
    #         / | \            /   \
    #        /  |  \          /     \   If tri1.dominantPixel is the same as
    #       /tri|tri\   ===> /_______\  from tri2.dominantPixel ...
    #       \ 1 | 2 /        \       /
    #        \  |  /          \     /
    #         \ | /            \   /
    #          \|/              \ /
    #                            

    # If itchy and scratchy are pinned, the process should be aborted
    if itchy.pinned() and scratchy.pinned():
        return []

    tri2nodes = tri2.nodes
    t0 = tri2nodes[(tri2nodes.index(itchy) + 1) % 3]

    change0 = skeleton.ProvisionalChanges()
    change0.moveNode(anchor, 0.5*(itchy.position() + scratchy.position()))
    change0.removeElements(tri1, tri2)
    parents = tri2.getParents()
    change0.insertElements(
        ProvisionalTriangle([anchor, t0, scratchy], parents=parents),
        ProvisionalTriangle([anchor, itchy, t0], parents=parents))

    change1 = skeleton.ProvisionalChanges()
    change1.removeElements(tri1, tri2)
    parents = tri1.getParents() + tri2.getParents()
    change1.insertElements(
        ProvisionalTriangle([anchor, t0, scratchy], parents=parents),
        ProvisionalTriangle([anchor, itchy, t0], parents=parents))

    # If anchor is limited in movility, change0 is not possible.
    if not anchor.movable_x() or not anchor.movable_y():
        return [change1]
    else:
        return [change0, change1]
                     
##############################################################################

def triNoneSplit(skel, anchor, itchy, scratchy, tri):
    #                                  
    # scratchy /|                |
    #         / |                |
    #        /  |              A |
    #   ____/tri|   ===>     ----|
    # anchor\   |                |
    #        \  |              B |
    #         \ |                |
    #   itchy  \|                |
    #

    # If itchy or scratchy is at the corner of skeleton, the process should be
    # aborted.
    if not itchy.movable_x() and not itchy.movable_y():
        return []
    if not scratchy.movable_x() and not scratchy.movable_y():
        return []
    # If either of three nodes is pinned, the process should be aborted.
    if anchor.pinned(): return []
    if itchy.pinned(): return []
    if scratchy.pinned(): return []
    
    change = skeleton.ProvisionalChanges()
    change.moveNode(anchor,
                    position=0.5*(itchy.position() + scratchy.position()),
                    mobility=(itchy.movable_x() or scratchy.movable_x(),
                              itchy.movable_y() or scratchy.movable_y()))
    change.removeElements(tri)
    return [change]
