// -*- C++ -*-
// $RCSfile: shapefunction.C,v $
// $Revision: 1.6 $
// $Author: reida $
// $Date: 2003/04/02 14:52:33 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include "element.h"
#include "masterelement.h"
#include "shapefunction.h"
#include "shapefunctioncache.h"
#include "common/tostring.h"
#include "common/trace.h"
using namespace std;

class ShapeFunctionTable {
  // Stores values of shape function and its derivatives at a set of
  // Gauss points.
private:
  ShapeFunctionTable(int ngauss, int nnodes);
  // f[i][j] = f(gausspoint i, node j)
  vector<vector<double> > f_table;
  // df[i][j][k] = df(gausspoint i, node j)/dx_k
  vector<vector<vector<double> > > df_table;
  friend class ShapeFunction;
};

ShapeFunctionTable::ShapeFunctionTable(int ngauss, int nsf)
  : f_table(ngauss, vector<double>(nsf)),
    df_table(ngauss, vector<vector<double> >(nsf, vector<double>(spacedim)))
{
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

ShapeFunction::ShapeFunction(int nsf, const MasterElement &master)
  : sftable(master.ngauss_sets()),
    sfcache(master.ngauss_sets()),
    nfunctions(nsf)
{
  for(int i=0; i<master.ngauss_sets(); i++) {
    sftable[i] = new ShapeFunctionTable(master.ngauss(i), nsf);
    sfcache[i] = new ShapeFunctionCache(master.ngauss(i), nsf);
  }
}

ShapeFunction::~ShapeFunction() {
  for(int i=0; i<sftable.size(); i++)
    delete sftable[i]; 
  for(int i=0; i<sfcache.size(); i++)
    delete sfcache[i];

}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Use double dispatch to evaluate shape functions at Positions, since
// the evaluation is done differently at GaussPoints and MasterCoords

double ShapeFunction::value(ShapeFunctionIndex n, const MasterPosition &p)
  const
{
  //  Trace("ShapeFunction::value p=" + to_string(p.mastercoord()));
  return p.shapefunction(*this, n);
}

double ShapeFunction::masterderiv(ShapeFunctionIndex n, SpaceIndex j,
				  const MasterPosition &p)
  const
{
  //  Trace("ShapeFunction::masterderiv sf=" + to_string(n) + " p=" + to_string(p.mastercoord()));
  return p.mdshapefunction(*this, n, j);
}

double ShapeFunction::realderiv(const Element *el, ShapeFunctionIndex n,
				SpaceIndex j, const MasterPosition &p)
  const
{
  //  Trace("ShapeFunction::realderiv sf=" + to_string(n) + " p=" + to_string(p.mastercoord()));
  return p.dshapefunction(el, *this, n, j);
}

// Find the value and derivative at Gauss points by using the lookup tables.

double ShapeFunction::value(ShapeFunctionIndex n, const GaussPoint &g)
  const
{
  return sftable[g.order()]->f_table[g.index()][n];
}

// derivative wrt master coordinates
double ShapeFunction::masterderiv(ShapeFunctionIndex n, SpaceIndex j,
			    const GaussPoint &g) const
{
  //  Trace("ShapeFunction::masterderiv sf=" + to_string(n) + " gpt=" + to_string(g.mastercoord()));
  return sftable[g.order()]->df_table[g.index()][n][j];
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Fill in the lookup tables

void ShapeFunction::precompute(const MasterElement &master) {
  // Use the virtual functions that evaluate the shapefunction at
  // arbitrary points to store its values at the Gauss points.

  // loop over integration orders (sets of gauss points)
  for(int ord=0; ord<master.ngauss_sets(); ord++) {

    const GaussPtTable &gptable = master.gptable(ord);

    sftable[ord] = new ShapeFunctionTable(gptable.size(), nfunctions);
    vector<vector<double> > &f_table = sftable[ord]->f_table;
    vector<vector<vector<double> > > &df_table = sftable[ord]->df_table;

    for(int g=0; g<gptable.size(); g++) { // loop over gausspoints
      for(ShapeFunctionIndex n=0; n<nfunctions; ++n) { // loop over sf's
	f_table[g][n] = value(n, gptable[g].position);
	for(SpaceIndex j=0; j<spacedim; ++j) // loop over spatial dimensions
	  df_table[g][n][j] = this->masterderiv(n, j, gptable[g].position);
      }
    }
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

double ShapeFunction::realderiv(const Element *el,
				ShapeFunctionIndex n, SpaceIndex i,
				const GaussPoint &g) const
{
  //  Trace("ShapeFunction::realderiv 1");
  double result = 0;
  if(sfcache[g.order()]->query_dsf(el, n, i, g, result))
    return result;

  // don't be tempted to rewrite this in terms of
  // realderiv(Element*, ..., MasterCoord&) because that one doesn't use
  // the precomputed values of the shape function derivatives!
  for(SpaceIndex j=0; j<spacedim; ++j)
    result += el->Jdmasterdx(j, i, g)*masterderiv(n, j, g);
  result /= el->det_jacobian(g);

  sfcache[g.order()]->store_dsf(el, n, i, g, result);
  return result;
}

double ShapeFunction::realderiv(const Element *el,
				ShapeFunctionIndex n, SpaceIndex i,
				const MasterCoord &mc) const
{
  //  Trace("ShapeFunction::realderiv 2");
  double result = 0;
  for(SpaceIndex j=0; j<spacedim; ++j)
    result += el->Jdmasterdx(j, i, mc)*masterderiv(n, j, mc);
  result /= el->det_jacobian(mc);
  return result;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

double ShapeFunction::det_jacobian(const Element *el, const GaussPoint &g) const
{
  double result;
  if(sfcache[g.order()]->query_jac(el, g, result))
    return result;

  // don't be tempted to rewrite this in terms of
  // det_jacobian(Element*, MasterCoord&) because that one doesn't use
  // the precomputed values of the shape function derivatives!
  result = el->jacobian(0, 0, g) * el->jacobian(1, 1, g) -
    el->jacobian(0, 1, g) * el->jacobian(1, 0, g);

  sfcache[g.order()]->store_jac(el, g, result);
  return result;
}

double ShapeFunction::det_jacobian(const Element *el, const MasterCoord &mc)
  const
{
  // This is wrong in three dimensions
  return el->jacobian(0, 0, mc) * el->jacobian(1, 1, mc) -
    el->jacobian(0, 1, mc) * el->jacobian(1, 0, mc);
}


